clear all; close all; 
clc;
%MIT License, see LICENCE file

%Copyright (c) 2019 Yui Chun Leung
%Modified by Albert Argilaga, 2021, for the publication:
% Predicting the non-deterministic response of Poroelastic Media 
% with Damageable Cracks using Generative Adversarial Networks
%% Basic Generative Adversarial Network

%% Load Data
load('mnistAll.mat')
trainX = preprocess(epsfv2D(:,1:12000)); %2D plane
%trainX = preprocess(epsfv3D(:,1:4096)); %3D plane, needs changes im. sizes

%% Load new colormaps
m=100;
cm_magma=magma(m);

% figure(1) %uncomment to plot the training samples
% plot(-epsfv2D(1,:),-epsfv2D(2,:),'r+')
% xlabel('\epsilon_{xx}')
% ylabel('\epsilon_{yy}')
% title("Training samples")
% axis square
%% Settings (hyperparameters)
settings.latent_dim = 64;
settings.batch_size = 24; settings.image_size = [128,128,1]; 
settings.lrD = 0.0003; settings.lrG = 0.0003; settings.beta1 = 0.201;
settings.beta2 = 0.799; settings.maxepochs = 151;

%% Initialization
%% Generator
fct=32; %neurons cound (network width)
paramsGen.FCW1 = dlarray(...
    initializeGaussian([fct*1,settings.latent_dim],.01));
paramsGen.FCb1 = dlarray(zeros(fct*1,1,'single'));
paramsGen.BNo1 = dlarray(zeros(fct*1,1,'single'));
paramsGen.BNs1 = dlarray(ones(fct*1,1,'single'));
paramsGen.FCW2 = dlarray(initializeGaussian([fct*2,fct*1]));
paramsGen.FCb2 = dlarray(zeros(fct*2,1,'single'));
paramsGen.BNo2 = dlarray(zeros(fct*2,1,'single'));
paramsGen.BNs2 = dlarray(ones(fct*2,1,'single'));
%Next layer is inactive to reduce depth
% paramsGen.FCW3 = dlarray(initializeGaussian([fct*1024,fct*512]));
% paramsGen.FCb3 = dlarray(zeros(fct*1024,1,'single'));
% paramsGen.BNo3 = dlarray(zeros(fct*1024,1,'single'));
% paramsGen.BNs3 = dlarray(ones(fct*1024,1,'single'));
paramsGen.FCW4 = dlarray(initializeGaussian(...
    [prod(settings.image_size),fct*2]));%1024
paramsGen.FCb4 = dlarray(zeros(prod(settings.image_size)...
    ,1,'single'));

stGen.BN1 = []; stGen.BN2 = []; stGen.BN3 = [];

%% Discriminator
%Next layer is inactive to reduce depth
% paramsDis.FCW1 = dlarray(initializeGaussian([fct*1024,...
%      prod(settings.image_size)],.02));
% paramsDis.FCb1 = dlarray(zeros(fct*1024,1,'single'));
% paramsDis.BNo1 = dlarray(zeros(fct*1024,1,'single'));
% paramsDis.BNs1 = dlarray(ones(fct*1024,1,'single'));
%paramsDis.FCW2 = dlarray(initializeGaussian([fct*512,fct*512]));%512 1024
paramsDis.FCW2 = dlarray(initializeGaussian([fct*2,...
      prod(settings.image_size)],.02));
paramsDis.FCb2 = dlarray(zeros(fct*2,1,'single'));
paramsDis.BNo2 = dlarray(zeros(fct*2,1,'single'));
paramsDis.BNs2 = dlarray(ones(fct*2,1,'single'));
paramsDis.FCW3 = dlarray(initializeGaussian([fct*1,fct*2]));
paramsDis.FCb3 = dlarray(zeros(fct*1,1,'single'));
paramsDis.FCW4 = dlarray(initializeGaussian([1,fct*1]));
paramsDis.FCb4 = dlarray(zeros(1,1,'single'));

stDis.BN1 = []; stDis.BN2 = [];

% average Gradient and average Gradient squared holders
avgG.Dis = []; avgGS.Dis = []; avgG.Gen = []; avgGS.Gen = [];
%% Train
numIterations = floor(size(trainX,2)/settings.batch_size)
%numIterations = 5;
out = false; epoch = 0; global_iter = 0;
while ~out
    tic; 
    trainXshuffle = trainX(:,randperm(size(trainX,2)));
    %fprintf('Epoch %d\n',epoch) 
    for i=1:numIterations
        global_iter = global_iter+1;
        noise = gpdl(randn([settings.latent_dim,...
            settings.batch_size]),'CB');
        idx = (i-1)*settings.batch_size+1:i*settings.batch_size;
        XBatch=gpdl(single(trainXshuffle(:,idx)),'CB');

        [GradGen,GradDis,stGen,stDis] = ...
                dlfeval(@modelGradients,XBatch,noise,...
                paramsGen,paramsDis,stGen,stDis);

        % Update Discriminator network parameters
        [paramsDis,avgG.Dis,avgGS.Dis] = ...
            adamupdate(paramsDis, GradDis, ...
            avgG.Dis, avgGS.Dis, global_iter, ...
            settings.lrD, settings.beta1, settings.beta2);

        % Update Generator network parameters
        [paramsGen,avgG.Gen,avgGS.Gen] = ...
            adamupdate(paramsGen, GradGen, ...
            avgG.Gen, avgGS.Gen, global_iter, ...
            settings.lrG, settings.beta1, settings.beta2);

        if i==1 || rem(i,20)==0
            progressplot(paramsGen,stGen,settings,epoch);
             if i==1 || (epoch>=0 && i==1) 
                h = gcf;
                % Capture the plot as an image 
                frame = getframe(h); 
                im = frame2im(frame); 
                [imind,cm] = rgb2ind(im,256); 
                % Write to the GIF File 
                if 1 % A.A. print .GIF (change to zero to avoid printing)
                    if epoch == 0
                      imwrite(imind,cm,'GANmnist.gif','gif', 'Loopcount',inf); 
                    else 
                      imwrite(imind,cm,'GANmnist.gif','gif','WriteMode','append'); 
                    end 
                end
            end
        end
               
    end

    elapsedTime = toc;
    disp("Epoch "+epoch+". Time taken for epoch = "+elapsedTime + "s")
    epoch = epoch+1;
    if epoch == settings.maxepochs
        out = true;
    end    
end
%% Helper Functions
%% preprocess
function x = preprocess(x)
x(1,:) = -x(1,:)+ min (x(1,:));
x(2,:) = x(2,:);

%x = double(x)/255;
%x = (x-.5)/.5;
%x = reshape(x,64*64,[]);

rows = 128; %A. Argilaga: function of number of steps used in the \mu-scale
columns = 128; % bis
%grayLevels = mat2gray(z(:))
% Rescale x and y
% Convert scaled x and y to integers with round() because they are row and column
% indexes, which need to be integers.
x(1,:) = round(rescale(x(1,:), 1, columns));
x(2,:) = round(rescale(x(2,:), 1, rows));
dataim=500 % this is the number of samples to build a training image
nimg=fix(length(x(1,:))/dataim)
t=0;
    grayImage = zeros(rows, columns, nimg);
for img=1:nimg
    for k = 1 : dataim %28*28=784
        % Get the row and column for this y and x, respectively.
        t=t+1;
        row = x(2,t);
        col = x(1,t);
        % Now assign the gray level at that location.
        grayImage(row, col, img) = grayImage(row, col, img)+1/25;
    end
    %imshow(grayImage(:,:,img), []);
end
%A.A. plot one real image
%I = rescale(grayImage(:,:,1));

% A.A.: this is for visualization of the real probability pool
%I=grayImage(:,:,1)+grayImage(:,:,2)+grayImage(:,:,3)+grayImage(:,:,4)+...
%    grayImage(:,:,5)+grayImage(:,:,6)+grayImage(:,:,7)+grayImage(:,:,8)+...
%    grayImage(:,:,9)+grayImage(:,:,10)+grayImage(:,:,11)+grayImage(:,:,12);
%I = rescale(I);

% figure(2) %uncomment to plot the real probability map
% %subplot(1,2,1)
% imagesc(I)
% title("Real probability map 1")
% colormap magma
% axis square
% colorbar
% 
% drawnow;

t
x = reshape(grayImage,128*128,[]);
size(x)

end
%% extract data
function x = gatext(x)
x = gather(extractdata(x));
end
%% gpu dl array wrapper
function dlx = gpdl(x,labels)
dlx = gpuArray(dlarray(x,labels));
end
%% Weight initialization
function parameter = initializeGaussian(parameterSize,sigma)
if nargin < 2
    sigma = 0.05;
end
parameter = randn(parameterSize, 'single') .* sigma;
end
%% Generator
function [dly,st] = Generator(dlx,params,st)
% fully connected
%1
dly = fullyconnect(dlx,params.FCW1,params.FCb1);
dly = leakyrelu(dly,0.2);
% if isempty(st.BN1)
%     [dly,st.BN1.mu,st.BN1.sig] = batchnorm(dly,params.BNo1,params.BNs1);
% else
%     [dly,st.BN1.mu,st.BN1.sig] = batchnorm(dly,params.BNo1,...
%         params.BNs1,st.BN1.mu,st.BN1.sig);
% end
%2
dly = fullyconnect(dly,params.FCW2,params.FCb2);
dly = leakyrelu(dly,0.2);
% if isempty(st.BN2)
%     [dly,st.BN2.mu,st.BN2.sig] = batchnorm(dly,params.BNo2,params.BNs2);
% else
%     [dly,st.BN2.mu,st.BN2.sig] = batchnorm(dly,params.BNo2,...
%         params.BNs2,st.BN2.mu,st.BN2.sig);
% end
%3
%dly = fullyconnect(dly,params.FCW3,params.FCb3);
%dly = leakyrelu(dly,0.2);
%  if isempty(st.BN3)
%      [dly,st.BN3.mu,st.BN3.sig] = batchnorm(dly,params.BNo3,params.BNs3);
%  else
%      [dly,st.BN3.mu,st.BN3.sig] = batchnorm(dly,params.BNo3,...
%          params.BNs3,st.BN3.mu,st.BN3.sig);
%  end
%4
dly = fullyconnect(dly,params.FCW4,params.FCb4);
% tanh
dly = tanh(dly);
end
%% Discriminator
function [dly,st] = Discriminator(dlx,params,st)
% fully connected 
%1
%dly = fullyconnect(dlx,params.FCW1,params.FCb1);
%dly = leakyrelu(dly,0.2);
%dly = dropout(dly);
% if isempty(st.BN1)
%     [dly,st.BN1.mu,st.BN1.sig] = batchnorm(dly,params.BNo1,params.BNs1);
% else
%     [dly,st.BN1.mu,st.BN1.sig] = batchnorm(dly,params.BNo1,...
%         params.BNs1,st.BN1.mu,st.BN1.sig);
% end
%2
dly = fullyconnect(dlx,params.FCW2,params.FCb2); %(dly,params.FCW2,params.FCb2);
dly = leakyrelu(dly,0.2);
dly = dropout(dly);
% if isempty(st.BN2)
%     [dly,st.BN2.mu,st.BN2.sig] = batchnorm(dly,params.BNo2,params.BNs2);
% else
%     [dly,st.BN2.mu,st.BN2.sig] = batchnorm(dly,params.BNo2,...
%         params.BNs2,st.BN2.mu,st.BN2.sig);
% end
%3
dly = fullyconnect(dly,params.FCW3,params.FCb3);
dly = leakyrelu(dly,0.2);
dly = dropout(dly);
%4
dly = fullyconnect(dly,params.FCW4,params.FCb4);
% sigmoid
dly = sigmoid(dly);
end
%% modelGradients
function [GradGen,GradDis,stGen,stDis]=modelGradients(x,z,paramsGen,...
    paramsDis,stGen,stDis)
[fake_images,stGen] = Generator(z,paramsGen,stGen);
d_output_real = Discriminator(x,paramsDis,stDis);
[d_output_fake,stDis] = Discriminator(fake_images,paramsDis,stDis);

% Loss due to true or not
d_loss = -mean(.9*log(d_output_real+eps)+log(1-d_output_fake+eps));
g_loss = -mean(log(d_output_fake+eps));

% For each network, calculate the gradients with respect to the loss.
GradGen = dlgradient(g_loss,paramsGen,'RetainData',true);
GradDis = dlgradient(d_loss,paramsDis);
end
%% progressplot
function progressplot(paramsGen,stGen,settings,epoch)
r = 4; c = 4;
noise = gpdl(randn([settings.latent_dim,r*c]),'CB');
gen_imgs = Generator(noise,paramsGen,stGen);
gen_imgs = reshape(gen_imgs,128,128,[]);

fig = gcf;
if ~isempty(fig.Children)
    delete(fig.Children)
end

I = imtile(gatext(gen_imgs));
I = rescale(I);
%figure(2)
%subplot(1,2,2)
imagesc(I)
title(['Generated probability maps (4x4), Epoch = ' num2str(epoch)])
colormap magma
axis square
colorbar

drawnow;
end
%% dropout
function dly = dropout(dlx,p)
if nargin < 2
    p = .3;
end
n = p*10;
mask = randi([1,10],size(dlx));
mask(mask<=n)=0;
mask(mask>n)=1;
dly = dlx.*mask;

end